module Main where

import Data.Char
import Data.List
import qualified Data.Map as M
import Debug.Trace

type Db = M.Map String Node

data Node = Node {
    big :: Bool
  , nbs :: [String]
  } deriving Show

type Path = [String]

main :: IO ()
main = do
  raw <- getContents
  let input = mkDb $ map splitInp $ lines raw
    in do
      putStrLn $ "day12a: " ++ (show $ solveA input)
      putStrLn $ "day12b: " ++ (show $ solveB input)

mkDb :: [(String,String)] -> Db
mkDb adjs = foldl f (M.fromList $ (map mkInitDb $ getUniques adjs)) adjs
  where
    f db (a,b) =
      M.adjust (\n -> n { nbs = (b:(nbs n)) }) a
      $ M.adjust (\n -> n { nbs = (a:(nbs n)) }) b
      $ db

mkInitDb :: String -> (String,Node)
mkInitDb s = (,) s
  $ Node (isUpper (head s)) []

getUniques :: [(String,String)] -> [String]
getUniques is = nub $ foldl (\t (a,b) -> (a:b:t)) [] is

findPaths :: (Db -> [String] -> [[String]]) -> [[String]] -> [[String]] -> Db -> [[String]]
findPaths f [] comp _ = comp
findPaths f (k:ks) comp db = findPaths f (ks ++ new) (nk comp) db
  where
    nk = if (head k) == "end"
         then (k:)
         else id
    new = if (head k) /= "end"
          then f db k
          else []

nextSteps :: Db -> [String] -> [[String]]
nextSteps db me = filter f $ map (\n -> (n:me)) nobs
  where
    f ss = (isUpper $ (head . head) ss)
         || not (any (\n -> n == (head ss)) (tail ss))
    nobs = (nbs $ db M.! (head me))

nextSteps' :: Db -> [String] -> [[String]]
nextSteps' db me = filter f $ map (\n -> (n:me)) nobs
  where
    f ss = (isUpper $ (head . head) ss)
         || length (g ss) == length (nub (g ss))+1
         || not (any (\n -> n == (head ss)) (tail ss))
    nobs = filter (/= "start") (nbs $ db M.! (head me))
    g ss = filter (\s -> not $ isUpper (head s)) ss

solveA :: Db -> Int
solveA db = length $ findPaths nextSteps [["start"]] [] db

solveB :: Db -> Int
solveB db = length $ findPaths nextSteps' [["start"]] [] db

-- io and parsing

splitInp :: String -> (String,String)
splitInp s = (\(a,b) -> (a, drop 1 b)) $ splitAt n s
  where
    n = findIn '-' s

findIn :: Char -> String -> Int
findIn c s = go 0 c s
  where
    go n c (s:ss) = if c == s then n else go (n+1) c ss
    go n c [] = error "bad parse"
